#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "compression.h"

#if 0
uint8_t search[] = {
0x80, 0x00, 0x00, 0x00, 0x48, 0x00, 0x01, 0xdd, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x20, 0x00, 0x00, 0x00, 0x0a, 0x01, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0xff, 0x80, 0x8d, 0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0x00, 0x20, 0x00, 0x00, 0x57, 0xc0, 0x00, 0x00,
0xff, 0x80, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x57, 0xfd, 0xf5, 0x30,
0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x8d, 0x88, 0x00, 0x2a, 0x01, 0x10, 0xff, 0x97, 0xdf, 0xa8,
0xff, 0x80, 0xdf, 0x50, 0xff, 0x80, 0xa6, 0xe0, 0xff, 0x81, 0x0a, 0xb8, 0xff, 0x81, 0xe8, 0x18,
0xff, 0x81, 0xb9, 0x68, 0xff, 0x81, 0x3d, 0x20, 0xff, 0x81, 0xe6, 0x10, 0xff, 0x81, 0xe3, 0x50,
0xff, 0x83, 0x9b, 0x98, 0xff, 0x83, 0x9c, 0x30, 0x00, 0x00, 0x00, 0x00, 0xff, 0xf8, 0x01, 0x84,
0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0xff, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x58, 0x00, 0x00, 0x00, 0x00, 0x01, 0x10, 0x00, 0x53, 0x72, 0x4e, 0x00,
0x05, 0xf0, 0x3e, 0x4d, 0x01, 0x7c, 0x0f, 0x93, 0x01, 0xfc, 0xa0, 0x55, 0x00, 0x00, 0x00, 0x00,
};
const size_t searchlen = sizeof(search);
#else

#endif

FILE *infile;

static void usage() {
	fprintf(stderr, "Usage: lzss infile { --encode | [ --decode ] [ searchfor [ romoffset [ compressionstart ] ] ] }\n");
	exit(1);
}

#define shift() do { if (argc) {argc--; argv++;} } while(0)

int main(int argc, char **argv)
{
	int enc = 0;
	char *s;
	long romoffset = 0;
	long compressionstart = 0;
	const char *search = NULL;
	size_t searchlen = 0;

	shift(); // remove executable name

	if (!argc) {
		usage();
	}
	char *inpath = *argv;
	shift();

	if (argc) {
		if (!strcmp(*argv, "--encode")) {
			enc = 1; shift();
			if (argc)
				usage();
		}
		else if (!strcmp(*argv, "--decode")) {
			enc = 0; shift();
		}
	}

	if (argc) {
		search = *argv;
		searchlen = strlen(search);
		shift();
	}

	if (argc) {
		sscanf(*argv, "%lx", &romoffset);
		shift();
	}

	if (argc) {
		sscanf(*argv, "%lx", &compressionstart);
		shift();
	}

	if (argc) {
		usage();
	}

	//fprintf(stderr, "input:\"%s\" %s search:%s offset:%lx start:%lx\n", inpath, enc ? "--encode" :  "--decode", search, romoffset, compressionstart );

	if ((infile = fopen(inpath, "rb")) == NULL) {
		fprintf(stderr, "file not found \"%s\"\n", inpath);
		exit(1);
	}

	int fd = fileno(infile); // if you have a stream (e.g. from fopen), not a file descriptor.
	struct stat buf;
	fstat(fd, &buf);

	size_t srclen = buf.st_size;

	//fprintf(stderr, "reading %zu bytes\n", srclen);

	uint8_t *src = malloc(srclen);
	read(fd, src, srclen);
	fclose(infile);

	//fprintf(stderr, "decompressing\n");

	size_t dstlen;
	for (size_t srcoff = compressionstart; srcoff < srclen; srcoff++ ) {
		// fprintf(stderr, "src:%zu len:%zu", srcoff, srclen - srcoff);

		size_t dstmaxlen = srclen * 2;
		if (dstmaxlen < 1 * 1024 * 1024)
			dstmaxlen = 1 * 1024 * 1024;

		uint8_t *dst = NULL;
		while (!dst) {
			dst = malloc(dstmaxlen);
			if (!dst) {
				fprintf(stderr, "out of memory");
				exit(1);
			}

			if (enc) {
				uint8_t *dstend = compress_lzss(dst, dstmaxlen, src + srcoff, srclen - srcoff);
				dstlen = dstend - dst;
			}
			else {
				dstlen = decompress_lzss(dst, dstmaxlen, src + srcoff, srclen - srcoff);
			}
			if (dstlen >= dstmaxlen) {
				free(dst);
				dst = NULL;
				dstmaxlen *= 2;
			}
		}
		// fprintf(stderr, " dstlen:%zu\n", dstlen);

		if (search && search[0]) {
			uint8_t *found = memmem(dst, dstlen, search, searchlen);
			if (found) {
				fprintf(stderr, "found rom:0x%08lx -> OF:0x%08lx len:0x%08lx\n", romoffset + srcoff, romoffset + found - dst, dstlen);
			}
		}
		else {
			fwrite(dst, dstlen, 1, stdout);
			break;
		}
	}

	fprintf(stderr, "input: %zu bytes\noutput: %zu bytes (%lu%%)\n", srclen, dstlen, (dstlen * 100 + (srclen / 2)) / srclen);
	return 0;
}
